/*
  X11 interface for oomer.
  Copyright (C) 2010 Deekoo L.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#define _GNU_SOURCE
#include <unistd.h>
#include <fcntl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/keysym.h>
#include <sys/select.h>

Window win;
int win_mapped = 0;
static XSetWindowAttributes win_attributes;
int win_valuemask = CWWinGravity|CWBackPixel|CWOverrideRedirect;
Display *display;
char buf[1024];
int bufptr=0; // Position to write new bytes to.
int bufpos=0; // Position of first unread byte in the buffer.
int fatalerror = 0;
int bufsize = 4;
char **log_buf = NULL;
typedef enum {
        c_green, c_yellow, c_orange
        } log_color;
int log_buf_colors[4];
int numbufitems = 0;
Font glob_font;
GC glob_gc;
GC orange_gc;
GC yellow_gc;
GC black_gc;
static XWMHints win_hints;
Colormap colormap;

GC ColoredGC(const char *colorstr)
{
  GC ret = XCreateGC(display,win,0,0);
  XColor tmp;
  XParseColor(display, colormap, colorstr, &tmp);
  XAllocColor(display, colormap, &tmp);
  XSetForeground(display, ret, tmp.pixel);
  return ret;
}

void create_window(void)
{
  int w = XDisplayWidth(display,0);
  int h = XDisplayHeight(display,0);
  win = XCreateWindow(display, RootWindow(display, 0), w-500, h-80, 500, 80, 0, 
        CopyFromParent, /* depth */
        InputOutput, /* class */
        CopyFromParent, /* visual - what's that? */
        win_valuemask, /* valuemask */
        &win_attributes);
  XMapWindow(display, win);
  win_mapped = 1;
  colormap = DefaultColormap(display, 0);
  glob_gc = XCreateGC(display, win, 0, 0);
  yellow_gc = ColoredGC("#EEEE00");
  orange_gc = ColoredGC("#FF0000");
  black_gc = XCreateGC(display, win, 0, 0);
  XSetForeground(display, black_gc, BlackPixel(display,0));
  XColor green_col;
  XParseColor(display, colormap, "#00FF00", &green_col);
  XAllocColor(display, colormap, &green_col);
  XSetForeground(display, glob_gc, green_col.pixel);
  XSetBackground(display, glob_gc, BlackPixel(display, 0));
  XSetWMHints(display, win, &win_hints);
  XSelectInput(display, win, ExposureMask | KeyPressMask | StructureNotifyMask | EnterWindowMask);
  XFlush(display);
}

void draw_dismiss_hint(void)
{
  XDrawImageString(display, win, glob_gc, 350, 15, "[Press space to dismiss]", 24);
}

void redraw(void)
{
  if (!win_mapped) create_window();
  XFillRectangle(display, win, black_gc, 0, 0, 500, 80);
  XTextItem ti;
  int x = 0;
  int y = 15;
  while (x<numbufitems) {
    ti.chars = log_buf[x];
    ti.nchars = strlen(log_buf[x]);
    ti.delta = 1;
    ti.font = None;
    GC gc;
    switch (log_buf_colors[x]) {
      case c_green:
        gc = glob_gc;
        break;
      case c_yellow:
        gc = yellow_gc;
        break;
      case c_orange:
        gc = orange_gc;
        break;
      default:
        fprintf(stderr,"error: unexpected color code %d.\n",log_buf_colors[x]);
        gc = orange_gc;
    }
    XDrawText(display, win, gc, 5, y, &ti, 1);
    y += 15;
    x++;
  }
  XDrawRectangle(display, win, glob_gc, 1, 1, 498, 78);
  draw_dismiss_hint();
}

void handlekeypress(XEvent *ev)
{
  if (XLookupKeysym(&ev->xkey,0)==XK_space) {
    int x = 0;
    while (x<numbufitems) {
      if (log_buf_colors[x]==c_orange) log_buf_colors[x] = c_yellow;
      x++;
    }
    win_mapped = 0;
    XUnmapWindow(display,win);
  } else {
    fprintf(stderr,"Unexpected keysym %d\n",ev->xkey);
  }
  XFlush(display);
}

void oomer_log(log_color c, char *s)
{
  numbufitems++;
  if (bufsize<numbufitems) {
    free(log_buf[0]);
    memmove(&log_buf[0],&log_buf[1],(bufsize-1)*sizeof(char *));
    memmove(&log_buf_colors[0],&log_buf_colors[1], (bufsize-1)*sizeof(int));
    numbufitems--;
  }
  log_buf[numbufitems-1] = strdup(s);
  log_buf_colors[numbufitems-1] = c;
}

void XHandleEvents(int fd)
{
  XEvent ev;
  do {
    XNextEvent(display, &ev);
    switch (ev.type) {
      case Expose:
        redraw();
        XFlush(display);
        break;
      case DestroyNotify:
        break;
      case EnterNotify:
        XSetInputFocus(display, win, RevertToPointerRoot, CurrentTime);
        break;
      case UnmapNotify:
        win_mapped = 0;
        break;
      case KeyPress:
        handlekeypress(&ev);
        break;
      case MapNotify:
        if (win_mapped) {
          win_mapped = 2;
          redraw();
          XFlush(display);
        }
        break;
      case ReparentNotify:
      case ConfigureNotify:
        fprintf(stderr, "An apparently unimportant notify %d\n",ev.type);
        break;
      default:
        fprintf(stderr,"Unexpected event %d\n",ev.type);
    }
  } while (XPending(display));
}

int main(int *argc, char **argv)
{
  log_buf = malloc(bufsize*sizeof(char *));
  if (!buf) return -1;
  display = XOpenDisplay(NULL);
  int xfd = XConnectionNumber(display);
  if (!display) {
    fprintf(stderr,"Error opening display.\n");
    exit(1);
  }
  oomer_log(c_green,"Starting oomer.");
  win_attributes.win_gravity = SouthEastGravity;
  win_attributes.background_pixel = BlackPixel(display,0);
  win_hints.flags = InputHint|XUrgencyHint;
  win_hints.input = 1;
  win_attributes.override_redirect = 1;
  create_window();
  glob_font = XLoadFont(display,"-*-fixed-bold-r-*-*-13-*-*-*-*-*-*-1");
  XEvent ev;
  int pipefds[2];
  if (pipe2(pipefds,O_NONBLOCK)) {
    fatalerror = 1;
    oomer_log(c_orange,"Error creating pipe.");
  }
  int highfd = pipefds[0];
  if (xfd>pipefds[0]) highfd = xfd;
  highfd++;
  fd_set rfds;
  FD_ZERO(&rfds);
  FD_SET(xfd,&rfds);
  FD_SET(pipefds[0],&rfds);
  pid_t child = fork();
  if (!child) {
    close(pipefds[0]);
    dup2(pipefds[1],1);
    close(0);
    execl("oomer","oomer",NULL);
  } else {
    close(pipefds[1]);
  }
  redraw();
  XFlush(display);
  while (1) {
    fd_set readable = rfds;
    if (select(highfd, &readable, NULL, NULL, NULL)>0) {
      if (FD_ISSET(pipefds[0],&readable)) {
        if (bufpos) {
          if (bufpos<bufptr) {
            memmove(&buf[0],&buf[bufpos], bufptr-bufpos);
            bufptr -= bufpos;
          } else {
            bufptr = 0;
          }
          bufpos = 0;
        }
        int nread = read(pipefds[0],&buf[bufptr],1024-bufptr);
        if (nread>0) {
          bufptr += nread;
        } else {
          fatalerror = 1;
          oomer_log(c_orange,"Read error on child pipe.");
          close(pipefds[0]);
          FD_CLR(pipefds[0],&rfds);
          waitpid(child, WNOHANG);
          redraw();
          XFlush(display);
        }
        while (bufptr-bufpos>=16) {
          int *ip = (int *)(&buf[bufpos]);
          bufpos += 16;
          if (ip[0]==*((int *)"KILL") && ip[3]==*((int *)"...\n")) {
            char procbuf[1024];
            if (snprintf(&procbuf[0], 128, "/proc/%d/cmdline", ip[1])) {
              int f = open(&procbuf[0],O_RDONLY);
              if (f>=0) {
                char procname[128];
                if (read(f,&procname,128)>=0) {
                  procname[127] ='\0';
                  snprintf(&procbuf[0],1024,"Killed %d [%s] with signal %d",ip[1],procname,ip[2]);
                  procbuf[1023] = '\0';
                  oomer_log(c_orange,&procbuf[0]);
                } else {
                  snprintf(&procbuf[0],1024,
                    "Killed %d <no cmdline> with signal %d", ip[1],ip[2]);
                  oomer_log(c_orange, &procbuf[0]);
                }
                close(f);
              } else {
                snprintf(&procbuf[0],1024,
                  "Killed %d <no cmdline> with signal %d", ip[1],ip[2]);
                oomer_log(c_orange, &procbuf[0]);
              }
            }
          } else {
            fprintf(stderr, "Error parsing input\n");
            oomer_log(c_orange, "Error parsing input.");
          }
          redraw();
          XFlush(display);
        }
      }
      if (FD_ISSET(xfd, &readable)) {
        XHandleEvents(xfd);
      }
    }
  }
  return 0;
}
